/*
    source.c
    Copyright (C) 2000  Kh. Naba Kumar Singh

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* Most of the code stolen from Glade and modified */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include <gnome.h>
#include "anjuta.h"
#include "utilities.h"
#include "source.h"

/*************************************************************************
 * Build Files.
 *************************************************************************/
static gboolean
source_write_module_file_list (ProjectDBase * data, FILE * fp, gboolean with_line_break, gchar * prefix,
			       PrjModule module)
{
	GList *files, *node;
	gboolean first;
	
	g_return_val_if_fail (data != NULL, FALSE);
	g_return_val_if_fail (fp != NULL, FALSE);

	files = project_dbase_get_module_files (data, module);
	first = TRUE;
	node = files;
	while (node)
	{
		gchar *str;
		
		if (node->data)
		{
			if (prefix)
				str = g_strconcat (prefix, "/", node->data, NULL);
			else
				str = g_strdup (node->data);
			if (with_line_break)
			{
				/* We avoid printing the first '\' in the file */
				if (first)
					fprintf (fp, "\t%s", str);
				else
					fprintf (fp, "\\\n\t%s", str);
			}
			else
			{
				fprintf (fp, "%s\n", str);
			}
			first = FALSE;
			g_free (str);
		}
		node = g_list_next (node);
	}
	fprintf (fp, "\n\n");
	glist_strings_free (files);
	return TRUE;
}


static gboolean
source_write_module_object_file_list (ProjectDBase * data, FILE * fp, gboolean with_line_break, gchar * prefix,
			       PrjModule module)
{
	GList *files, *node;
	gboolean first;
	
	g_return_val_if_fail (data != NULL, FALSE);
	g_return_val_if_fail (fp != NULL, FALSE);

	files = project_dbase_get_module_files (data, module);
	first = TRUE;
	node = files;
	while (node)
	{
		gchar *str;
		
		if (node->data)
		{
			gchar** ofile_array;
			gchar* ofile;
			
			ofile_array = g_strsplit(node->data,".", 2);
			ofile = g_strconcat(ofile_array[0], ".o",NULL);
			g_strfreev(ofile_array);
			
			if (prefix)
				str = g_strconcat (prefix, "/", ofile, NULL);
			else
				str = g_strdup (ofile);
			
			g_free(ofile);
			
			if (with_line_break)
			{
				/* We avoid printing the first '\' in the file */
				if (first)
					fprintf (fp, "\t%s", str);
				else
					fprintf (fp, "\\\n\t%s", str);
			}
			else
			{
				fprintf (fp, "%s\n", str);
			}
			first = FALSE;
			g_free (str);
		}
		node = g_list_next (node);
	}
	fprintf (fp, "\n\n");
	glist_strings_free (files);
	return TRUE;
}




/* This creates an empty file in the given directory if it doesn't already
   exist. It is used to create the empty NEWS, README, AUTHORS & ChangeLog. */
static gboolean
source_create_file_if_not_exist (const gchar * directory,
				 const gchar * filename,
				 const gchar * contents)
{
	gchar *pathname;
	FILE *fp;
	gint bytes_written, ret;

	g_return_val_if_fail (directory != NULL, FALSE);
	g_return_val_if_fail (filename != NULL, FALSE);

	ret = TRUE;
	pathname = g_strconcat (directory, "/", filename, NULL);

	if (!file_is_regular (pathname))
	{
		fp = fopen (pathname, "w");
		if (fp)
		{
			if (contents)
			{
				gint contents_len;
				contents_len = strlen (contents);
				bytes_written = fwrite (contents, 1, contents_len, fp);
				if (bytes_written != contents_len)
				{
					anjuta_system_error (errno, _("Error writing to file: %s."), pathname);
					ret = FALSE;
				}
			}
			fclose (fp);
		}
		else
		{
			anjuta_system_error (errno, _("Unable to create file: %s."), filename);
			ret = FALSE;
		}
	}
	g_free (pathname);
	return ret;
}

static gboolean
source_write_no_modify_warning (ProjectDBase * data, FILE * fp)
{
	g_return_val_if_fail (data != NULL, FALSE);
	g_return_val_if_fail (fp != NULL, FALSE);

	fprintf (fp,
	"# Created by Anjuta-TIa - will be overwritten\n"
	"# If you don't want it to overwrite it,\n"
	"# \tPlease disable it in the Anjuta-TIa project configuration\n\n");
	return TRUE;
}

/* Copies the generic autogen.sh script which runs aclocal, automake
   & autoconf to create the Makefiles etc. */
static gboolean
source_write_autogen_sh (ProjectDBase * data)
{
	gchar *srcbuffer, *destbuffer;
	gint old_umask;
	Project_Type* type;

	g_return_val_if_fail (data != NULL, FALSE);

	type = project_dbase_get_project_type(data);

	srcbuffer = g_strconcat (app->dirs->data, type->autogen_file, NULL);
	destbuffer = g_strconcat (data->top_proj_dir, "/autogen.sh", NULL);

	/* FIXME: If *.desktop.in exists, just leave it, for now. */
	if (file_is_regular (destbuffer))
	{
		g_free (destbuffer);
		g_free (srcbuffer);
		return TRUE;
	}

	if (!copy_file (srcbuffer, destbuffer, FALSE))
	{
		anjuta_system_error (errno, _("Unable to create file: %s."), destbuffer);
		g_free (destbuffer);
		g_free (srcbuffer);
		return FALSE;
	}

	/* We need to make the script executable, but we try to honour any umask. */
	old_umask = umask (0666);
	chmod (destbuffer, 0777 & ~old_umask);
	umask (old_umask);

	g_free (srcbuffer);
	g_free (destbuffer);
	return TRUE;
}

static gboolean
source_write_toplevel_makefile (ProjectDBase * data)
{
	FILE *fp;
	gchar *filename, *actual_file, *target, *prj_name;
	gint i;
	gchar* prjname;
	gchar* version;
	Project_Type* type;

	g_return_val_if_fail (data != NULL, FALSE);

	filename = get_a_tmp_file();
	actual_file = g_strconcat (data->top_proj_dir, "/Makefile", NULL);
	type = project_dbase_get_project_type(data);
	fp = fopen (filename, "w");
	if (fp == NULL)
	{
		anjuta_system_error (errno, _("Unable to create file: %s."), filename);
		g_free (filename);
		g_free (actual_file);
		return FALSE;
	}

	fprintf (fp, "# Top Level Makefile\n");
	source_write_no_modify_warning (data, fp);

	fprintf (fp, "SUBDIRS =");

	for (i = 0; i < MODULE_END_MARK; i++)
	{
		gchar *subdir;
		if (i == MODULE_PO)
			continue;
		if (project_dbase_module_is_empty (data, i))
			continue;

		subdir = project_dbase_get_module_name (data, i);

		if (subdir)
		{
			fprintf (fp, " %s", subdir);
			g_free (subdir);
		}
	}
	
	fprintf (fp, "\n\n");

	fprintf(fp,
		"all:\n"
		"\tfor i in $(SUBDIRS); do \\\n"
		"\tcd $$i; \\\n"
		"\t$(MAKE); \\\n"
		"\tcd -; \\\n"
		"\tdone\n");
	
	fprintf (fp, "\n\n");

	fprintf(fp,
		"clean:\n"
		"\tfor i in $(SUBDIRS); do \\\n"
		"\tcd $$i; \\\n"
		"\t$(MAKE) clean; \\\n"
		"\tcd -; \\\n"
		"\tdone\n");

	fprintf (fp, "\n\n");

	prjname = project_dbase_get_proj_name (data);
	version = project_dbase_get_version (data);
	
	fprintf(fp,
		"dist:\n"
		"\tfor i in $(SUBDIRS); do \\\n"
		"\tcd $$i; \\\n"
		"\t$(MAKE) clean; \\\n"
		"\tcd -; \\\n"
		"\tdone;\\\n"
		"\ttar cvzf %s-%s.tar.gz .", target, version);

	g_free(target);
	g_free(version);

	fprintf (fp, "\n\n");

	/*
	prj_name = project_dbase_get_proj_name (data);
	target = prop_get (data->props, "project.source.target");
	g_strdelimit (target, "-", '_');

	fprintf (fp,
		"%sdocdir = ${prefix}/doc/%s\n"
		"%sdoc_DATA = \\\n"
		"\tREADME\\\n"
		"\tCOPYING\\\n"
		"\tAUTHORS\\\n"
		"\tChangeLog\\\n"
		"\tINSTALL\\\n"
		"\tNEWS\\\n"
		"\tTODO", target, prj_name, target);
	

	fprintf (fp,
		"\n\nEXTRA_DIST = $(%sdoc_DATA)\n\n", target);
	*/
	
	/*fprintf (fp, "# Copy all the spec files. Of cource, only one is actually used.\n");
	fprintf (fp, "dist-hook:\n"
		  "\tfor specfile in *.spec; do \\\n"
		  "\t\tif test -f $$specfile; then \\\n"
		  "\t\t\tcp -p $$specfile $(distdir); \\\n"
		  "\t\tfi \\\n"
		  "\tdone\n"
		  "\n");
		  */
	fclose (fp);
	if (move_file_if_not_same(filename, actual_file) == FALSE)
		anjuta_system_error (errno, _("Unable to create file: %s."), actual_file);
	g_free (actual_file);
	g_free (target);
	g_free (prj_name);
	g_free (filename);
	free_project_type(type);
	return TRUE;
}

static gboolean
source_write_po_files (ProjectDBase * data)
{
	gchar *dirname = NULL, *filename = NULL, *actual_file, *prefix, *str;
	FILE *fp;

	g_return_val_if_fail (data != NULL, FALSE);

	return TRUE;

	str = project_dbase_get_module_name (data, MODULE_PO);
	dirname = g_strconcat (data->top_proj_dir, "/", str, NULL);
	g_free (str);
	force_create_dir (dirname);

	/* Create ChangeLog if it doesn't exist. */
	source_create_file_if_not_exist (dirname, "/ChangeLog", NULL);

	filename = get_a_tmp_file();
	actual_file  = g_strconcat (dirname, "/POTFILES.in", NULL);
	g_free (dirname);

	fp = fopen (filename, "w");
	if (fp == NULL)
	{
		anjuta_system_error (errno, _("Unable to create file: %s."), filename);
		g_free (filename);
		g_free (actual_file);
		return FALSE;
	}

	/* We need the relative path from the project directory to the source
	 * directory, to prefix each source file. */
	prefix = project_dbase_get_module_name (data, MODULE_SOURCE);
	fprintf (fp, "# List of source files containing translatable strings.\n\n");
	source_write_no_modify_warning (data, fp);
	fprintf (fp,"# Source files\n");
	source_write_module_file_list (data, fp, FALSE, prefix, MODULE_SOURCE);
	fclose (fp);
	if (move_file_if_not_same(filename, actual_file) == FALSE)
		anjuta_system_error (errno, _("Unable to create file: %s."), actual_file);
	g_free (actual_file);
	g_free (prefix);
	g_free (filename);
	return TRUE;
}

static gboolean
source_write_executable_source_files (ProjectDBase * data)
{
	FILE *fp;
	gchar *src_dir, *filename, *actual_file, *target;
	Project_Type* type;
	
	g_return_val_if_fail (data != NULL, FALSE);

	type = project_dbase_get_project_type(data);

	if (project_dbase_module_is_empty (data, MODULE_SOURCE))
		return TRUE;
	
	src_dir = project_dbase_get_module_dir (data, MODULE_SOURCE);
	force_create_dir (src_dir);
	g_free (src_dir);

	src_dir = project_dbase_get_module_name (data, MODULE_SOURCE);
	filename = get_a_tmp_file();
	actual_file =	g_strconcat (data->top_proj_dir, "/", src_dir, "/Makefile", NULL);
	g_free (src_dir);

	fp = fopen (filename, "w");
	if (fp == NULL)
	{
		anjuta_system_error (errno, _("Unable to create file: %s."), filename);
		g_free (filename);
		free_project_type(type);
		g_free (actual_file);
		return FALSE;
	}

	fprintf (fp,"# Makefile for executable target\n\n");
	source_write_no_modify_warning (data, fp);
	/* If the project directory is the source directory, we need to output
	 * SUBDIRS here. */
	/* Fixed bug #460321 "Problem with optimization options" by separating 
	   INCLUDE AND C(XX)FLAGS. Johannes Schmid 3.2.2002 */ 
	
	fprintf (fp, "INCLUDES =");
	fprintf (fp, type->cflags);
	compiler_options_set_prjincl_in_file (app->compiler_options, fp);
	fprintf (fp, "\n\n");
	
	
	fprintf (fp, "CFLAGS =");
	compiler_options_set_prjcflags_in_file (app->compiler_options, fp);
	fprintf (fp, "\n\n");
	
	target =
		prop_get (data->props, "project.source.target");
	g_strdelimit (target, "-", '_');

/*	fprintf (fp, "bin_PROGRAMS = %s\n\n", target);
	fprintf (fp, "%s_SOURCES = \\\n", target);*/

	fprintf (fp, "OBJ = \\\n");
//	source_write_module_file_list (data, fp, TRUE, NULL, MODULE_SOURCE);
	source_write_module_object_file_list(data, fp, TRUE, NULL, MODULE_SOURCE);

	fprintf (fp, "LDFLAGS = ");
	compiler_options_set_prjlflags_in_file (app->compiler_options, fp);
	fprintf (fp, "\n\n");
	
	fprintf (fp, "LDADD =");
	fprintf (fp, type->ldadd);
	compiler_options_set_prjlibs_in_file (app->compiler_options, fp);
	fprintf (fp, "\n\n");

	fprintf(fp,"TMP = tempfile.o\n\n");
	fprintf(fp,"all:dummy\n\n");
	fprintf(fp,"dummy: $(TMP)\n"
			   "\tobj2ti $^ -o %s\n\n",target);
	fprintf(fp, "$(TMP):$(OBJ)\n"
				"\tm68k-coff-ld $(LDFLAGS) -r -e 0 $(OBJ) $(LDADD) -o $@\n\n");
	
	fprintf(fp,"%%.o:%%.c\n"
			   "\ttigcc $(INCLUDES) $(CFLAGS) -c $^\n\n");
	fprintf(fp,"%%.o:%%.s\n"
			   "\ttigcc $(INCLUDES) $(CFLAGS) -c $^\n\n");
	fprintf(fp,"%%.o:%%.asm\n"
			   "\ttigcc $(INCLUDES) $(CFLAGS) -c $^\n\n");		   
	fprintf(fp,"clean:\n"
			   "\t-rm *.o *.??[yz]\n\n");

	fclose (fp);
	
	if (move_file_if_not_same(filename, actual_file) == FALSE)
		anjuta_system_error (errno, _("Unable to create file: %s."), actual_file);
	
	g_free (actual_file);
	g_free (target);
	free_project_type(type);
	g_free (filename);
	return TRUE;
}

static gboolean
source_write_static_lib_source_files (ProjectDBase * data)
{
	FILE *fp;
	gchar *src_dir, *filename, *actual_file, *target;
	Project_Type* type;
	
	g_return_val_if_fail (data != NULL, FALSE);

	type = project_dbase_get_project_type(data);

	if (project_dbase_module_is_empty (data, MODULE_SOURCE))
		return TRUE;
	
	src_dir = project_dbase_get_module_dir (data, MODULE_SOURCE);
	force_create_dir (src_dir);
	g_free (src_dir);

	src_dir = project_dbase_get_module_name (data, MODULE_SOURCE);
	filename = get_a_tmp_file();
	actual_file =	g_strconcat (data->top_proj_dir, "/", src_dir, "/Makefile", NULL);
	g_free (src_dir);

	fp = fopen (filename, "w");
	if (fp == NULL)
	{
		anjuta_system_error (errno, _("Unable to create file: %s."), filename);
		g_free (filename);
		free_project_type(type);
		g_free (actual_file);
		return FALSE;
	}

	fprintf (fp,"# Makefile for static lib target\n\n");
	source_write_no_modify_warning (data, fp);
	/* If the project directory is the source directory, we need to output
	 * SUBDIRS here. */
	/* Fixed bug #460321 "Problem with optimization options" by separating 
	   INCLUDE AND C(XX)FLAGS. Johannes Schmid 3.2.2002 */ 
	
	fprintf (fp, "INCLUDES =");
	fprintf (fp, type->cflags);
	compiler_options_set_prjincl_in_file (app->compiler_options, fp);
	fprintf (fp, "\n\n");
		
	fprintf (fp, "CFLAGS =");
	compiler_options_set_prjcflags_in_file (app->compiler_options, fp);
	fprintf (fp, "\n");

	target = prop_get (data->props, "project.source.target");
	g_strdelimit (target, "-", '_');

	/* automake says that this is invalid */
	fprintf (fp, "LDFLAGS = ");
	compiler_options_set_prjlflags_in_file (app->compiler_options, fp);
	fprintf (fp, "\n\n");
	
	fprintf (fp, "LDADD = ");
	fprintf (fp, type->ldadd);
	compiler_options_set_prjlibs_in_file (app->compiler_options, fp);
	fprintf (fp, "\n\n");
		
	fprintf (fp, "OBJ = \\\n");
//	source_write_module_file_list (data, fp, TRUE, NULL, MODULE_SOURCE);
	source_write_module_object_file_list(data, fp, TRUE, NULL, MODULE_SOURCE);

	fprintf (fp, "all:%s%s\n\n", target, ".a");
	fprintf (fp, "%s%s: $(OBJ)\n" 
				 "\tm68k-coff-ar -r $@ $(OBJ) $(LDADD)\n\n"
				, target, ".a");
	fprintf(fp, "%%.o:%%.c\n"
			   "\ttigcc $(INCLUDES) $(CFLAGS) -c $^\n\n");
	fprintf(fp,"%%.o:%%.s\n"
			   "\ttigcc $(INCLUDES) $(CFLAGS) -c $^\n\n");
	fprintf(fp,"%%.o:%%.asm\n"
			   "\ttigcc $(INCLUDES) $(CFLAGS) -c $^\n\n");		   
	fprintf(fp,"clean:\n"
			   "\t-rm *.[ao] *.??[yz]\n\n");			
	
	
	fclose (fp);
	
	if (move_file_if_not_same(filename, actual_file) == FALSE)
		anjuta_system_error (errno, _("Unable to create file: %s."), actual_file);
	
	g_free (actual_file);
	g_free (target);
	free_project_type(type);
	g_free (filename);
	return TRUE;
}

static gboolean
source_write_dynamic_lib_source_files (ProjectDBase * data)
{
	FILE *fp;
	gchar *src_dir, *filename, *actual_file, *target;
	Project_Type* type;
	
	g_return_val_if_fail (data != NULL, FALSE);

	type = project_dbase_get_project_type(data);

	if (project_dbase_module_is_empty (data, MODULE_SOURCE))
		return TRUE;
	
	src_dir = project_dbase_get_module_dir (data, MODULE_SOURCE);
	force_create_dir (src_dir);
	g_free (src_dir);

	src_dir = project_dbase_get_module_name (data, MODULE_SOURCE);
	filename = get_a_tmp_file();
	actual_file =	g_strconcat (data->top_proj_dir, "/", src_dir, "/Makefile.am", NULL);
	g_free (src_dir);

	fp = fopen (filename, "w");
	if (fp == NULL)
	{
		anjuta_system_error (errno, _("Unable to create file: %s."), filename);
		g_free (filename);
		free_project_type(type);
		g_free (actual_file);
		return FALSE;
	}

	fprintf (fp,"# Makefile for dynamic lib target\n\n");
	source_write_no_modify_warning (data, fp);
	/* If the project directory is the source directory, we need to output
	 * SUBDIRS here. */
	/* Fixed bug #460321 "Problem with optimization options" by separating 
	   INCLUDE AND C(XX)FLAGS. Johannes Schmid 3.2.2002 */ 
	
	fprintf (fp, "INCLUDES =");
	fprintf (fp, type->cflags);
	compiler_options_set_prjincl_in_file (app->compiler_options, fp);
	fprintf (fp, "\n\n");
	
		fprintf (fp, "CFLAGS =");
		compiler_options_set_prjcflags_in_file (app->compiler_options, fp);
		fprintf (fp, "\n\n");
	
	target =
		prop_get (data->props, "project.source.target");
	g_strdelimit (target, "-", '_');
	fprintf (fp, "lib_LTLIBRARIES = %s%s\n\n", target, ".la");
	fprintf (fp, "%s_%s_SOURCES = \\\n", target, "la");
	source_write_module_file_list (data, fp, TRUE, NULL, MODULE_SOURCE);

	fprintf (fp, "%s_%s_LDFLAGS = ", target, "la");
	/* If any equivalent for this could be done in type->... */
	/* fprintf (fp, type->ldadd); */
	compiler_options_set_prjlflags_in_file (app->compiler_options, fp);
	fprintf (fp, "\n\n");
	
	fprintf (fp, "%s_%s_LIBADD = ", target, "la");
	fprintf (fp, type->ldadd);
	compiler_options_set_prjlibs_in_file (app->compiler_options, fp);
	fprintf (fp, "\n\n");
	fclose (fp);
	
	if (move_file_if_not_same(filename, actual_file) == FALSE)
		anjuta_system_error (errno, _("Unable to create file: %s."), actual_file);
	
	g_free (actual_file);
	g_free (target);
	free_project_type(type);
	g_free (filename);
	return TRUE;
}

static gboolean
source_write_source_files (ProjectDBase * data)
{
	gint target_type, ret;
	g_return_val_if_fail (data != NULL, FALSE);

	target_type = project_dbase_get_target_type (data);
	switch (target_type)
	{
		case PROJECT_TARGET_TYPE_EXECUTABLE:
			ret = source_write_executable_source_files (data);
			break;
		case PROJECT_TARGET_TYPE_STATIC_LIB:
			ret = source_write_static_lib_source_files (data);
			break;
		case PROJECT_TARGET_TYPE_DYNAMIC_LIB:
			ret = source_write_dynamic_lib_source_files (data);
			break;
		default:
			anjuta_error ("Unknown Project's target type");
			return FALSE;
	}
	return ret;
}

static gboolean
source_write_include_files (ProjectDBase * data)
{
	FILE *fp;
	gchar *src_dir, *filename, *actual_file, *target;
	g_return_val_if_fail (data != NULL, FALSE);

	if (project_dbase_module_is_empty (data, MODULE_INCLUDE))
		return TRUE;

	src_dir = project_dbase_get_module_dir (data, MODULE_INCLUDE);
	force_create_dir (src_dir);
	g_free (src_dir);

	src_dir = project_dbase_get_module_name (data, MODULE_INCLUDE);
	filename = get_a_tmp_file();
	actual_file =	g_strconcat (data->top_proj_dir, "/", src_dir, "/Makefile", NULL);
	g_free (src_dir);

	fp = fopen (filename, "w");
	if (fp == NULL)
	{
		anjuta_system_error (errno, _("Unable to create file: %s."), filename);
		g_free (filename);
		g_free (actual_file);
		return FALSE;
	}

	fprintf (fp,"# Makefile for include file\n\n");
	source_write_no_modify_warning (data, fp);

	target =
		prop_get (data->props, "project.source.target");
	g_strdelimit (target, "-", '_');
	/*fprintf (fp, "%s_includedir = $(pkgincludedir)\n\n", target);
	fprintf (fp, "%s_include_DATA = \\\n", target);
	source_write_module_file_list (data, fp, TRUE, NULL, MODULE_INCLUDE);
*/
	/*fprintf (fp, "EXTRA_DIST = $(%s_include_DATA)\n", target);*/
	fclose (fp);
	if (move_file_if_not_same(filename, actual_file) == FALSE)
		anjuta_system_error (errno, _("Unable to create file: %s."), actual_file);
	g_free (actual_file);
	g_free (target);
	g_free (filename);
	return TRUE;
}

static gboolean
source_write_pixmap_files (ProjectDBase * data)
{
	FILE *fp;
	gchar *src_dir, *filename, *actual_file, *target;
	Project_Type* type;

	g_return_val_if_fail (data != NULL, FALSE);

	type = project_dbase_get_project_type(data);

	if (project_dbase_module_is_empty (data, MODULE_PIXMAP))
		return TRUE;

	src_dir = project_dbase_get_module_dir (data, MODULE_PIXMAP);
	force_create_dir (src_dir);
	g_free (src_dir);

	src_dir = project_dbase_get_module_name (data, MODULE_PIXMAP);
	filename = get_a_tmp_file();
	actual_file =	g_strconcat (data->top_proj_dir, "/", src_dir, "/Makefile", NULL);
	g_free (src_dir);

	fp = fopen (filename, "w");
	if (fp == NULL)
	{
		anjuta_system_error (errno, _("Unable to create file: %s."), filename);
		g_free (filename);
		g_free (actual_file);
		free_project_type(type);
		return FALSE;
	}
	fprintf (fp,"## Process this file with automake to produce Makefile.in\n\n");
	source_write_no_modify_warning (data, fp);

	target = prop_get (data->props, "project.source.target");
	g_strdelimit (target, "-", '_');
	fprintf (fp, "%s_pixmapsdir = $(prefix)/@NO_PREFIX_PACKAGE_PIXMAPS_DIR@\n\n", target);
	
	fprintf (fp, "%s_pixmaps_DATA = \\\n", target);
	source_write_module_file_list (data, fp, TRUE, NULL, MODULE_PIXMAP);

	fprintf (fp, "EXTRA_DIST = $(%s_pixmaps_DATA)\n", target);
	fclose (fp);
	if (move_file_if_not_same(filename, actual_file) == FALSE)
		anjuta_system_error (errno, _("Unable to create file: %s."), actual_file);
	g_free (actual_file);
	g_free (target);
	g_free (filename);
	free_project_type(type);
	return TRUE;
}

static gboolean
source_write_help_files (ProjectDBase * data)
{
	FILE *fp;
	gchar *src_dir, *filename, *actual_file, *target;
	Project_Type* type;

	g_return_val_if_fail (data != NULL, FALSE);

	type = project_dbase_get_project_type(data);
	if (project_dbase_module_is_empty (data, MODULE_HELP))
		return TRUE;

	src_dir = project_dbase_get_module_dir (data, MODULE_HELP);
	force_create_dir (src_dir);
	g_free (src_dir);

	src_dir = project_dbase_get_module_name (data, MODULE_HELP);
	filename = get_a_tmp_file();
	actual_file =	g_strconcat (data->top_proj_dir, "/", src_dir, "/Makefile.am", NULL);
	g_free (src_dir);

	fp = fopen (filename, "w");
	if (fp == NULL)
	{
		anjuta_system_error (errno, _("Unable to create file: %s."), filename);
		g_free (filename);
		free_project_type(type);
		g_free (actual_file);
		return FALSE;
	}

	fprintf (fp,"## Process this file with automake to produce Makefile.in\n\n");
	source_write_no_modify_warning (data, fp);

	target = prop_get (data->props, "project.source.target");
	g_strdelimit (target, "-", '_');
	fprintf (fp, "%s_helpdir = $(prefix)/@NO_PREFIX_PACKAGE_HELP_DIR@/C\n\n", target);
	fprintf (fp, "%s_help_DATA = \\\n", target);
	source_write_module_file_list (data, fp, TRUE, NULL, MODULE_HELP);

	fprintf (fp, "EXTRA_DIST = $(%s_help_DATA)\n", target);
	fclose (fp);
	if (move_file_if_not_same(filename, actual_file) == FALSE)
		anjuta_system_error (errno, _("Unable to create file: %s."), actual_file);
	g_free (actual_file);
	g_free (target);
	free_project_type(type);
	g_free (filename);
	return TRUE;
}

static gboolean
source_write_doc_files (ProjectDBase * data)
{
	FILE *fp;
	gchar *src_dir, *filename, *actual_file, *target;

	g_return_val_if_fail (data != NULL, FALSE);

	if (project_dbase_module_is_empty (data, MODULE_DOC))
		return TRUE;

	src_dir = project_dbase_get_module_dir (data, MODULE_DOC);
	force_create_dir (src_dir);
	g_free (src_dir);

	src_dir = project_dbase_get_module_name (data, MODULE_DOC);
	filename = get_a_tmp_file();
	actual_file =	g_strconcat (data->top_proj_dir, "/", src_dir, "/Makefile.am", NULL);
	g_free (src_dir);

	fp = fopen (filename, "w");
	if (fp == NULL)
	{
		anjuta_system_error (errno, _("Unable to create file: %s."), filename);
		g_free (filename);
		g_free (actual_file);
		return FALSE;
	}
	fprintf (fp,"## Process this file with automake to produce Makefile.in\n\n");
	source_write_no_modify_warning (data, fp);

	target =
		prop_get (data->props, "project.source.target");
	g_strdelimit (target, "-", '_');
	fprintf (fp, "%s_docdir = $(prefix)/@NO_PREFIX_PACKAGE_DOC_DIR@\n\n", target);
	fprintf (fp, "%s_doc_DATA = \\\n", target);
	source_write_module_file_list (data, fp, TRUE, NULL, MODULE_DOC);

	fprintf (fp, "EXTRA_DIST = $(%s_doc_DATA)\n", target);
	fclose (fp);
	if (move_file_if_not_same(filename, actual_file) == FALSE)
		anjuta_system_error (errno, _("Unable to create file: %s."), actual_file);
	g_free (actual_file);
	g_free (target);
	g_free (filename);
	return TRUE;
}

static gboolean
source_write_data_files (ProjectDBase * data)
{
	FILE *fp;
	gchar *src_dir, *filename, *actual_file, *target;

	g_return_val_if_fail (data != NULL, FALSE);

	if (project_dbase_module_is_empty (data, MODULE_DATA))
		return TRUE;

	src_dir = project_dbase_get_module_dir (data, MODULE_DATA);
	force_create_dir (src_dir);
	g_free (src_dir);

	src_dir = project_dbase_get_module_name (data, MODULE_DATA);
	filename = get_a_tmp_file();
	actual_file =	g_strconcat (data->top_proj_dir, "/", src_dir, "/Makefile.am", NULL);
	g_free (src_dir);

	fp = fopen (filename, "w");
	if (fp == NULL)
	{
		anjuta_system_error (errno, _("Unable to create file: %s."), filename);
		g_free (filename);
		g_free (actual_file);
		return FALSE;
	}
	fprintf (fp,"## Process this file with automake to produce Makefile.in\n\n");
	source_write_no_modify_warning (data, fp);

	target =
		prop_get (data->props, "project.source.target");
	g_strdelimit (target, "-", '_');
	fprintf (fp, "%s_datadir = $(prefix)/@NO_PREFIX_PACKAGE_DATA_DIR@\n\n", target);
	fprintf (fp, "%s_data_DATA = \\\n", target);
	source_write_module_file_list (data, fp, TRUE, NULL, MODULE_DATA);

	fprintf (fp, "EXTRA_DIST = $(%s_data_DATA)\n", target);
	fclose (fp);
	if (move_file_if_not_same(filename, actual_file) == FALSE)
		anjuta_system_error (errno, _("Unable to create file: %s."), actual_file);
	g_free (actual_file);
	g_free (target);
	g_free (filename);
	return TRUE;
}

static void
source_write_platform_support(ProjectDBase *data, FILE *fp)
{
	gint tmp;
	
	tmp = prop_get_int(data->props, "project.has.platform.89.support",1);
	if (tmp == 1)
		fprintf(fp,
		"#define USE_TI89\t\t//Compile for TI89\n");

	tmp = prop_get_int(data->props, "project.has.platform.92p.support",1);
	if (tmp == 1)
		fprintf(fp,
		"#define USE_TI92PLUS\t\t//Compile for TI92 Plus\n");
	
	tmp = prop_get_int(data->props, "project.has.platform.v200.support",1);
	if (tmp == 1)
		fprintf(fp,
		"#define USE_V200\t\t//Compile for V200\n");	
	
}

static void
source_write_min_AMS_version(ProjectDBase* data, FILE* fp)
{
	gchar *str;
	str = prop_get(data->props, "project.min_AMS_version");
	if (str)
	{
		gdouble tmp;
		int minAMS;
		tmp = g_strtod (str, NULL);
		tmp *= 100;
		minAMS = (int)tmp;
		fprintf (fp, "#define MIN_AMS %d\n", minAMS);
		g_free (str);
	}
}

static void
source_write_dll_id_and_version(ProjectDBase* data, FILE* fp)
{
	gchar *str;
	gchar *major;
	gchar *minor;
	
	str = prop_get(data->props, "project.dll.id");
	if (str)
	{
		int dllId;
		dllId = strtol (str, NULL,10);
		fprintf (fp, "DLL_ID %d\n", dllId);
		g_free (str);
	}
	
	major = prop_get(data->props, "project.dll.major.version");
	minor = prop_get(data->props, "project.dll.minor.version");
	if (major && minor)
	{
		int dllMajor;
		int dllMinor;
		dllMajor = strtol (major, NULL,10);
		dllMinor = strtol (minor, NULL,10);
		fprintf (fp, "DLL_VERSION %d,%d\n", dllMajor, dllMinor);
		g_free (major);
		g_free (minor);
	}
}




static void
source_write_defines(ProjectDBase* data, FILE* fp)
{
	Project_Type* type;
	type = project_dbase_get_project_type(data);
	if (type->id == PROJECT_TYPE_NOSTUB)
	{
		gint tmp;
		tmp = prop_get_int(data->props, "project.has.ROM_calls_optimization",1);
		if (tmp == 1)
			fprintf(fp,
			"#define OPTIMIZE_ROM_CALLS\t\t//Use ROM calls optimization\n");	
	
		tmp = prop_get_int(data->props, "project.has.save_restore_LCD",1);
		if (tmp == 1)
			fprintf(fp,
			"#define SAVE_SCREEN\t\t//Save/Restore LCD contents\n");
	}
	free_project_type(type);	
}

gboolean
source_write_generic_main_c (ProjectDBase *data)
{

	FILE *fp;
	gchar *filename, *src_dir;
	Project_Type* type;
	gint tmp;
	gchar* str;
	
	g_return_val_if_fail (data != NULL, FALSE);
	g_return_val_if_fail (data->project_is_open, FALSE);

	
	src_dir = project_dbase_get_module_dir (data, MODULE_SOURCE);
	if (!src_dir)
		return FALSE;
	force_create_dir (src_dir);

	filename = g_strconcat (src_dir, "/main.c", NULL);
	
	g_free (src_dir);

	/* FIXME: If main.c exists, just leave it, for now. */
	if (file_is_regular (filename))
	{
		g_free (filename);
		return TRUE;
	}

	fp = fopen (filename, "w");
	if (fp == NULL)
	{
		anjuta_system_error (errno, _("Unable to create file: %s."), filename);
		g_free (filename);
		return FALSE;
	}
	
	type = project_dbase_get_project_type(data);

	fprintf(fp,
		"/* Created by Anjuta-TIa version %s */\n", VERSION);
	fprintf(fp,
		"/*\tThis file will not be overwritten */\n\n");
	
	if (type->id==PROJECT_TYPE_KERNELBASED)
	{
		fprintf(fp,"#define USE_KERNEL\t\t//include kernel support\n");
	}

	source_write_platform_support(data, fp);
	
	tmp = prop_get_int(data->props, "project.has.return_error_to_AMS",1);
	if (tmp == 1)
		fprintf(fp,
		"#define ENABLE_ERROR_RETURN\t\t//Enable returning errors to TIOS\n");	
	
	source_write_defines(data, fp);

	source_write_min_AMS_version(data,fp);	
	
	fprintf(fp,
		"#include <tigcclib.h>\n\n");
	
	if (type->id == PROJECT_TYPE_KERNELBASED)
	{
		str = prop_get(data->props, "project.doors.comment");
		if (str)
		{
			fprintf(fp, "char comment[] = \"%s\";\t\t//Doors program comment\n", str);
			g_free(str);
		}
	}
	
	fprintf(fp,		
		"void _main(void)\n"
		"{\n"
		"\tprintf(\"Hello world\\n\");\n"
		"}\n\n");
	
	fclose (fp);
	g_free (filename);
	free_project_type(type);
	return TRUE;
}


gboolean
source_write_static_lib_c (ProjectDBase *data)
{
	FILE *fp;
	gchar *filename, *src_dir;
	Project_Type* type;
	
	g_return_val_if_fail (data != NULL, FALSE);
	g_return_val_if_fail (data->project_is_open, FALSE);

	
	src_dir = project_dbase_get_module_dir (data, MODULE_SOURCE);
	if (!src_dir)
		return FALSE;
	force_create_dir (src_dir);

	filename = g_strconcat (src_dir, "/file.c", NULL);
	
	g_free (src_dir);

	/* FIXME: If main.c exists, just leave it, for now. */
	if (file_is_regular (filename))
	{
		g_free (filename);
		return TRUE;
	}

	fp = fopen (filename, "w");
	if (fp == NULL)
	{
		anjuta_system_error (errno, _("Unable to create file: %s."), filename);
		g_free (filename);
		return FALSE;
	}
	
	type = project_dbase_get_project_type(data);

	fprintf(fp,
		"/* Created by Anjuta-TIa version %s */\n", VERSION);
	fprintf(fp,
		"/*\tThis file will not be overwritten */\n\n");
	
	if (type->id==PROJECT_TYPE_KERNELBASED)
	{
		fprintf(fp,"#define USE_KERNEL\t\t//include kernel support\n");
	}

	source_write_platform_support(data, fp);
	
	/*tmp = prop_get_int(data->props, "project.has.return_error_to_AMS");
	if (tmp == 1)
		fprintf(fp,
		"#define ENABLE_ERROR_RETURN\t\t//Enable returning errors to TIOS\n");	
	*/
	source_write_defines(data, fp);

	source_write_min_AMS_version(data,fp);	
	
	fprintf(fp,
		"#include <tigcclib.h>\n\n");
	
	fprintf(fp,		
		"void function(void)\n"
		"{\n"
		"}\n\n");
	
	fclose (fp);
	g_free (filename);
	free_project_type(type);
	return TRUE;
}


gboolean
source_write_dynamic_lib_c (ProjectDBase *data)
{

	FILE *fp;
	gchar *filename, *src_dir;
	Project_Type* type;

	g_return_val_if_fail (data != NULL, FALSE);
	g_return_val_if_fail (data->project_is_open, FALSE);
	
	src_dir = project_dbase_get_module_dir (data, MODULE_SOURCE);
	if (!src_dir)
		return FALSE;
	force_create_dir (src_dir);

	filename = g_strconcat (src_dir, "/dll.c", NULL);
	
	g_free (src_dir);

	/* FIXME: If dll.c exists, just leave it, for now. */
	if (file_is_regular (filename))
	{
		g_free (filename);
		return TRUE;
	}

	fp = fopen (filename, "w");
	if (fp == NULL)
	{
		anjuta_system_error (errno, _("Unable to create file: %s."), filename);
		g_free (filename);
		return FALSE;
	}
	
	type = project_dbase_get_project_type(data);
	
	fprintf(fp,
		"/* Created by Anjuta-TIa version %s */\n", VERSION);
	fprintf(fp,
		"/*\tThis file will not be overwritten */\n\n");
	
	if (type->id == PROJECT_TYPE_NOSTUB)
	{
		fprintf(fp,
			"DLL_INTERFACE\n"
			"//declare here global vars and functions you want to export\n");
		source_write_dll_id_and_version(data, fp);
		fprintf(fp,
			"DLL_EXPORTS //Export here your symbols\n\n"
			"DLL_IMPLEMENTATION\n\n"
			"//Implement here your functions\n"
		);	
	}
	else if (type->id == PROJECT_TYPE_KERNELBASED)
	{
		fprintf(fp,
			"#include <stdio.h>\n\n"
			"int _library\n\n"
			"int mydll()\n"
			"{\n"
			"\treturn (0);\n"
			"}\n\n");					
	}
		
	fclose (fp);
	g_free (filename);
	free_project_type(type);
	return TRUE;
}


gboolean
source_write_build_files (ProjectDBase * data)
{
	Project_Type* type;
	gint ret;

	g_return_val_if_fail (data != NULL, FALSE);
	g_return_val_if_fail (data->project_is_open, FALSE);

	/* project is blocked, don't bother to disturb the
	project */
	if (data->project_config->blocked)
		return TRUE;
	
	ret = source_write_autogen_sh (data);
	if (!ret) return FALSE;
	
	type = project_dbase_get_project_type (data);
	free_project_type(type);

	if (data->project_config->disable_overwrite[BUILD_FILE_TOP_MAKEFILE] == FALSE)
	{
		ret = source_write_toplevel_makefile (data);
		if (!ret) return FALSE;
	}
	if (data->project_config->disable_overwrite[BUILD_FILE_SOURCE] == FALSE)
	{
		ret = source_write_source_files (data);
		if (!ret) return FALSE;
	}
	if (data->project_config->disable_overwrite[BUILD_FILE_INCLUDE] == FALSE)
	{
		ret = source_write_include_files (data);
		if (!ret) return FALSE;
	}
	if (data->project_config->disable_overwrite[BUILD_FILE_HELP] == FALSE)
	{
		ret = source_write_help_files (data);
		if (!ret) return FALSE;
	}
	if (data->project_config->disable_overwrite[BUILD_FILE_PIXMAP] == FALSE)
	{
		ret = source_write_pixmap_files (data);
		if (!ret) return FALSE;
	}
	if (data->project_config->disable_overwrite[BUILD_FILE_DATA] == FALSE)
	{
		ret = source_write_data_files (data);
		if (!ret) return FALSE;
	}
	if (data->project_config->disable_overwrite[BUILD_FILE_DOC] == FALSE)
	{
		ret = source_write_doc_files (data);
		if (!ret) return FALSE;
	}
	if (data->project_config->disable_overwrite[BUILD_FILE_PO] == FALSE)
	{
		ret = source_write_po_files (data);
		if (!ret) return FALSE;
	}

	ret = source_create_file_if_not_exist (data->top_proj_dir, "TODO", NULL);
	if (!ret) return FALSE;
	ret = source_create_file_if_not_exist (data->top_proj_dir, "NEWS", NULL);
	if (!ret) return FALSE;
	ret = source_create_file_if_not_exist (data->top_proj_dir, "README", NULL);
	if (!ret) return FALSE;
	ret = source_create_file_if_not_exist (data->top_proj_dir, "AUTHORS", NULL);
	if (!ret) return FALSE;
	ret = source_create_file_if_not_exist (data->top_proj_dir, "ChangeLog", NULL);
	if (!ret) return FALSE;
	ret = source_create_file_if_not_exist (data->top_proj_dir, "stamp-h.in", "timestamp\n");
	if (!ret) return FALSE;
	return TRUE;
}
